import os
from typing import Callable, Generic, Literal, Optional, Sequence, TypeVar
from lightning import LightningModule, Trainer
import lightning.pytorch as pl
from lightning.pytorch.callbacks import Callback, BasePredictionWriter
import torch


def _get_save_path(trainer: Trainer) -> str:
    if len(trainer.loggers) > 0 and trainer.loggers[0].save_dir is not None:
        save_dir = trainer.loggers[0].save_dir
        name = trainer.loggers[0].name
        version = trainer.loggers[0].version
        version = version if isinstance(
            version, str) else f"version_{version}"
        return os.path.join(
            save_dir, str(name), version)
    else:
        return trainer.default_root_dir


class CUDAMemorySnapshotCallback(Callback):

    def __init__(self, save_path: str | None = None) -> None:
        self.save_path = save_path

    def setup(self, trainer: Trainer, pl_module: LightningModule, stage: str) -> None:
        torch.cuda.memory._record_memory_history()

    def teardown(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule", stage: str) -> None:
        if self.save_path is None:
            self.save_path = os.path.join(_get_save_path(
                trainer), f"memory_snapshot_rank_{trainer.global_rank}.pt")
        torch.cuda.memory._dump_snapshot(self.save_path)


T = TypeVar('T')
T2 = TypeVar('T2')


def identity_transform(x: T) -> T:
    return x


class PredictionWriterCallback(BasePredictionWriter, Generic[T, T2]):
    def __init__(self, write_interval: Literal["batch", "epoch", "batch_and_epoch"] = "batch", save_path: str | None = None, transform_batch: Callable[[T], T] = identity_transform, transform_outputs: Callable[[T2], T2] = identity_transform) -> None:
        super().__init__(write_interval=write_interval)
        self.save_path = save_path
        self.transform_batch = transform_batch
        self.transform_outputs = transform_outputs

    def setup(self, trainer: pl.Trainer, pl_module: pl.LightningModule, stage: str) -> None:
        super().setup(trainer, pl_module, stage)
        if self.save_path is None:
            self.save_path = os.path.join(_get_save_path(
                trainer), f"outputs_rank_{trainer.global_rank}")
            os.makedirs(self.save_path, exist_ok=True)

    def write_on_batch_end(self, trainer: pl.Trainer,
                           pl_module: pl.LightningModule,
                           prediction: T2,
                           batch_indices: Optional[Sequence[int]],
                           batch: T,
                           batch_idx: int,
                           dataloader_idx: int,) -> None:
        torch.save(self.transform_outputs(prediction), os.path.join(self.save_path,  # type: ignore
                   f"outputs_{dataloader_idx}_{batch_idx}.pt"))
        torch.save(self.transform_batch(batch), os.path.join(self.save_path,  # type: ignore
                   f"batch_{dataloader_idx}_{batch_idx}.pt"))



import logging
import os
import sqlite3
from typing import Callable, Generic, Iterable, Iterator, Optional, Tuple, TypeVar, Sequence, TypeVarTuple, cast
from zipfile import ZipFile
import click
import fsspec
import torch.utils.data.dataloader
from torch.utils.data import Dataset, DataLoader
import torch
import lightning.pytorch
from edzip.sqlite import create_sqlite_directory_from_zip
from hscitorchutil.fsspec import get_s3fs_credentials, cache_locally_if_remote

Ts = TypeVarTuple("Ts")
T_co = TypeVar('T_co', covariant=True)
T2_co = TypeVar('T2_co', covariant=True)


def _identity(x: T_co) -> T_co:
    return x


class SQLiteDataset(Dataset[T_co], Generic[*Ts, T_co]):
    def __init__(self, sqlite_filename: str, table_name: str, index_column: str, columns_to_return: str, id_column: str):
        self.sqlite_filename = sqlite_filename
        self.table_name = table_name
        self.index_column = index_column
        self.id_column = id_column
        self.columns_to_return = columns_to_return
        self.sqlite = sqlite3.connect(sqlite_filename)
        self._len = self.sqlite.execute(
            f"SELECT COUNT(*) FROM {table_name}").fetchone()[0]

    def __len__(self):
        return self._len

    def __getitem__(self, idx: int | str) -> T_co:
        return self.__getitems__([idx])[0]

    def __getitems__(self, idxs: Sequence[int | str]) -> Sequence[T_co]:
        if isinstance(idxs[0], int):
            return self.sqlite.execute(f"SELECT {self.columns_to_return} FROM {self.table_name} WHERE {self.index_column} IN (%s)" % ','.join(
                '?' * len(idxs)), idxs).fetchall()
        else:
            return self.sqlite.execute(f"SELECT {self.columns_to_return} FROM {self.table_name} WHERE {self.id_column} IN (%s)" % ','.join(
                '?' * len(idxs)), idxs).fetchall()

    def __setstate__(self, state):
        (
            self.sqlite_filename,
            self.table_name,
            self.index_column,
            self.columns_to_return,
            self.id_column,
            self._len
        ) = state
        self.sqlite = sqlite3.connect(self.sqlite_filename)

    def __getstate__(self) -> object:
        return (
            self.sqlite_filename,
            self.table_name,
            self.index_column,
            self.columns_to_return,
            self.id_column,
            self._len
        )


def _remove_nones_from_batch(batch: Sequence) -> Sequence:
    """Removes None values from batch. Used to recover from errors."""
    batch = list(filter(lambda x: x is not None, batch))
    if batch:
        try:
            return torch.utils.data.dataloader.default_collate(batch)
        except Exception as e:
            logging.exception("Failed to collate batch, returning empty batch")
            return ()
    logging.warn("Batch is empty")
    return ()


class TypedDataLoader(Iterable[T_co], DataLoader[T_co], Generic[T_co]):
    pass


class SQLiteDataModule(lightning.pytorch.LightningDataModule, Generic[*Ts, T_co, T2_co]):
    def __init__(self,
                 train_sqlite_url: str,
                 val_sqlite_url: str,
                 test_sqlite_url: str,
                 cache_dir: str,
                 table_name: str,
                 index_column: str,
                 columns_to_return: str,
                 id_column: str,
                 storage_options: dict = dict(),
                 batch_size: int = 64,
                 num_workers: int = 0,
                 train_transform: Callable[[
                     Dataset[tuple[*Ts]]], Dataset[T_co]] = _identity,
                 test_transform: Callable[[
                     Dataset[tuple[*Ts]]], Dataset[T_co]] = _identity,
                 prepare_data_per_node: bool = True,
                 collate_fn: Optional[Callable[[Sequence[T_co]], T2_co]] = _remove_nones_from_batch):
        self.train_sqlite_url = train_sqlite_url
        self.val_sqlite_url = val_sqlite_url
        self.test_sqlite_url = test_sqlite_url
        self.cache_dir = cache_dir
        self.storage_options = storage_options
        self.batch_size = batch_size
        self.num_workers = num_workers

        self.table_name = table_name
        self.index_column = index_column
        self.columns_to_return = columns_to_return
        self.id_column = id_column
        self.train_transform = train_transform
        self.test_transform = test_transform
        self.prepare_data_per_node = prepare_data_per_node
        self.collate_fn = collate_fn
        self.train_dataset = None
        self.val_dataset = None
        self.test_dataset = None
        super().__init__()

    def prepare_data(self):
        # download, IO, etc. Useful with shared filesystems
        # only called on 1 GPU/TPU in distributed

        # Ensure sqlite databases are downloaded
        cache_locally_if_remote(
            self.train_sqlite_url, storage_options=self.storage_options, cache_dir=self.cache_dir)
        cache_locally_if_remote(
            self.val_sqlite_url, storage_options=self.storage_options, cache_dir=self.cache_dir)
        cache_locally_if_remote(
            self.test_sqlite_url, storage_options=self.storage_options, cache_dir=self.cache_dir)

    def setup(self, stage: Optional[str] = None):
        if (stage is None or stage == "fit") and self.train_dataset is None:
            self.train_dataset = self.train_transform(SQLiteDataset(cache_locally_if_remote(
                self.train_sqlite_url, storage_options=self.storage_options, cache_dir=self.cache_dir),
                self.table_name,
                self.index_column,
                self.columns_to_return,
                self.id_column))
        if (stage is None or stage == "fit" or stage == "validate") and self.val_dataset is None:
            self.val_dataset = self.test_transform(SQLiteDataset(cache_locally_if_remote(
                self.val_sqlite_url, storage_options=self.storage_options, cache_dir=self.cache_dir),
                self.table_name,
                self.index_column,
                self.columns_to_return,
                self.id_column))
        if (stage is None or stage == "test") and self.test_dataset is None:
            self.test_dataset = self.test_transform(SQLiteDataset(cache_locally_if_remote(
                self.test_sqlite_url, storage_options=self.storage_options, cache_dir=self.cache_dir),
                self.table_name,
                self.index_column,
                self.columns_to_return,
                self.id_column))

    def train_dataloader(self) -> TypedDataLoader[T2_co]:
        if self.train_dataset is None:
            raise ValueError("Training dataset not available")
        return cast(TypedDataLoader[T2_co], DataLoader(self.train_dataset, shuffle=True, batch_size=self.batch_size, num_workers=self.num_workers, persistent_workers=self.num_workers > 0,
                                                       collate_fn=self.collate_fn, pin_memory=True))

    def val_dataloader(self) -> TypedDataLoader[T2_co]:
        if self.val_dataset is None:
            raise ValueError("Validation dataset not available")
        return cast(TypedDataLoader[T2_co], DataLoader(self.val_dataset, batch_size=self.batch_size, num_workers=self.num_workers, persistent_workers=self.num_workers > 0, collate_fn=self.collate_fn, pin_memory=True))

    def test_dataloader(self) -> TypedDataLoader[T2_co]:
        if self.test_dataset is None:
            raise ValueError("Test dataset not available")
        return cast(TypedDataLoader[T2_co], DataLoader(self.test_dataset, batch_size=self.batch_size, num_workers=self.num_workers, persistent_workers=self.num_workers > 0, collate_fn=self.collate_fn, pin_memory=True))

    def teardown(self, stage: str):
        # clean up state after the trainer stops, delete files...
        # called on every process in DDP
        pass


@click.command()
@click.option("--key", required=False)
@click.option("--secret", required=False, help="AWS secret access key or file from which to read credentials")
@click.option("--endpoint-url", required=False)
@click.argument("zip-url")
@click.argument("sqlite-filename", required=True)
def main(zip_url: str, sqlite_filename: Optional[str] = None, endpoint_url: Optional[str] = None, key: Optional[str] = None, secret: Optional[str] = None):
    if secret is not None and os.path.exists(secret):
        credentials = get_s3fs_credentials(secret)
    else:
        credentials = dict()
    with fsspec.open(zip_url, **credentials) as zf:  # type: ignore
        create_sqlite_directory_from_zip(
            ZipFile(zf), sqlite_filename)  # type: ignore


if __name__ == "__main__":
    main()



import bisect
import random
from typing import Any, Callable, Optional, Sequence, TypeVar, Generic
import torch
from torch.utils.data import Dataset
import torch.utils.data
import logging

T_co = TypeVar('T_co', covariant=True)


class LinearMapSubset(Dataset[T_co], Generic[T_co]):
    r"""
    Slice a map dataset at specified indices.

    Args:
        dataset (Dataset[T_co]): The whole map dataset
        indices (sequence): Indices in the whole set selected for subset
    """
    dataset: Dataset[T_co]
    start: int
    end: int

    def __init__(self, dataset: Dataset[T_co], start: int = 0, end: Optional[int] = None) -> None:
        self.dataset = dataset
        self.start = start
        if end is not None:
            self.end = end
        else:
            self.end = len(self.dataset)  # type: ignore

    def __getitem__(self, idx):
        return self.dataset[self.start + idx]

    def __getitems__(self, indices: list[int]) -> list[T_co]:
        # add batched sampling support when parent dataset supports it.
        # see torch.utils.data._utils.fetch._MapDatasetFetcher
        if callable(getattr(self.dataset, "__getitems__", None)):
            return self.dataset.__getitems__([self.start + idx for idx in indices])  # type: ignore[attr-defined] # noqa
        else:
            return [self.dataset[self.start + idx] for idx in indices]

    def __len__(self):
        return self.end - self.start


T2_co = TypeVar('T2_co', covariant=True)


class TransformedMapDataset(Dataset[T2_co], Generic[T2_co, T_co]):
    r"""Create a transformed map dataset by applying a transform function to all samples.

    Args:
        dataset (Dataset[T_co]): The underlying map dataset
        transform (Callable[T:co,T2_co]): The transformation function to be applied to each sample
    """

    def __init__(self, dataset: Dataset[T_co], transform: Callable[[Sequence[T_co]], Sequence[T2_co]]) -> None:
        self.dataset = dataset
        self.transform = transform

    def __getitem__(self, idx):
        return self.transform([self.dataset[idx]])[0]

    def __getitems__(self, indices: Sequence[int]) -> Sequence[T2_co]:
        # add batched sampling support when parent dataset supports it.
        # see torch.utils.data._utils.fetch._MapDatasetFetcher
        if callable(getattr(self.dataset, "__getitems__", None)):
            return self.transform(self.dataset.__getitems__(indices))  # type: ignore[attr-defined] # noqa
        else:
            return self.transform([self.dataset[idx] for idx in indices])

    def __len__(self):
        return len(self.dataset)  # type: ignore


class ShuffledMapDataset(Dataset[T_co], Generic[T_co]):
    r"""
    Shuffle the input map dataset via its indices.

    Args:
        dataset (Dataset): Map dataset being shuffled
        seed: (int, optional): The seed to be used for shuffling. If not provided, the current time is used.
        indices (list[Any]): a list of indices for the parent Dataset. If not provided, we assume it uses 0-based indexing
    """
    dataset: Dataset[T_co]

    def __init__(self, dataset: Dataset[T_co], seed: int, indices: Optional[list[Any]] = None) -> None:
        self.dataset = dataset
        self.seed = seed
        self.indices = indices
        self._shuffle()

    def _shuffle(self):
        if self.indices is None:
            rng = torch.Generator().manual_seed(self.seed)
            self._shuffled_indices = torch.randperm(
                len(self.dataset), generator=rng).tolist()  # type: ignore
        else:
            rng = random.Random()
            rng.seed(self.seed)
            self._shuffled_indices: list = rng.sample(
                self.indices, len(self.indices))

    def __getitem__(self, idx):
        return self.dataset[self._shuffled_indices[idx]]

    def __getitems__(self, indices: list[int]) -> list[T_co]:
        # add batched sampling support when parent dataset supports it.
        # see torch.utils.data._utils.fetch._MapDatasetFetcher
        if callable(getattr(self.dataset, "__getitems__", None)):
            return self.dataset.__getitems__([self._shuffled_indices[idx] for idx in indices])  # type: ignore[attr-defined] # noqa
        else:
            return [self.dataset[self._shuffled_indices[idx]] for idx in indices]

    def __len__(self) -> int:
        return len(self.dataset)  # type: ignore

    def __getstate__(self):
        state = (
            self.dataset,
            self.indices,
            self.seed,
        )
        return state

    def __setstate__(self, state):
        (
            self.dataset,
            self.indices,
            self.seed,
        ) = state
        self._shuffle()


def _log_exception(ds: 'ExceptionHandlingMapDataset', idx: int, e: Exception) -> None:
    logging.exception(
        f"ExceptionHandlingMapDataset encountered exception at index {idx}. Returning None.")


class ExceptionHandlingMapDataset(Dataset[T_co], Generic[T_co]):
    r"""A dataset wrapper that catches exceptions and instead of bailing out, returns None.

    Args:
        dataset (Dataset[T_co]): The underlying map dataset
        on_exception (Callable[[int, Exception],Any]): The function to be called when an exception is raised.
    """

    def __init__(self, dataset: Dataset[T_co], on_exception: Callable[['ExceptionHandlingMapDataset', int, Exception], T_co] = _log_exception) -> None:
        self.dataset = dataset
        self.on_exception = on_exception

    def __getitem__(self, idx):
        try:
            return self.dataset[idx]
        except Exception as e:
            return self.on_exception(self, idx, e)

    def __getitems__(self, indices: list[int]) -> list[T_co]:
        # add batched sampling support when parent dataset supports it.
        # see torch.utils.data._utils.fetch._MapDatasetFetcher
        if callable(getattr(self.dataset, "__getitems__", None)):
            try:
                return self.dataset.__getitems__(indices)  # type: ignore[attr-defined] # noqa
            except Exception:
                return [self.__getitem__(idx) for idx in indices]
        else:
            return [self.__getitem__(idx) for idx in indices]  # type: ignore

    def __len__(self):
        return len(self.dataset)  # type: ignore


class DatasetToIterableDataset(torch.utils.data.IterableDataset[T_co], Generic[T_co]):
    def __init__(self, dataset: torch.utils.data.Dataset[T_co]):
        self.dataset = dataset

    def __iter__(self):
        if hasattr(self.dataset, "__iter__"):
            return self.dataset.__iter__()  # type: ignore
        for i in range(len(self.dataset)):  # type: ignore
            yield self.dataset[i]


class UnionMapDataset(Dataset[T_co], Generic[T_co]):
    def __init__(self, datasets: Sequence[Dataset[T_co]]) -> None:
        self.datasets = datasets
        self.supports_getitems = True
        start = 0
        self.start_offsets = []
        for dataset in datasets:
            if not callable(getattr(dataset, "__getitems__", None)):
                self.supports_getitems = False
            self.start_offsets.append(start)
            start += len(dataset)  # type: ignore
        self._len = start

    def __getitem__(self, idx):
        dataset_idx = bisect.bisect_right(self.start_offsets, idx) - 1
        return self.datasets[dataset_idx][idx - self.start_offsets[dataset_idx]]

    def __getitems__(self, indices: list[int]) -> list[T_co]:
        # add batched sampling support when parent dataset supports it.
        # see torch.utils.data._utils.fetch._MapDatasetFetcher
        if self.supports_getitems:
            idxs_by_datasets = [[] for _ in self.datasets]
            for idx_idx, idx in enumerate(indices):
                dataset_idx = bisect.bisect_right(self.start_offsets, idx) - 1
                idxs_by_datasets[dataset_idx].append(
                    (idx - self.start_offsets[dataset_idx], idx_idx))
            items = [None for _ in indices]
            for dataset_idx, idxs_by_dataset in enumerate(idxs_by_datasets):
                if idxs_by_dataset:
                    dataset_items = self.datasets[dataset_idx].__getitems__(  # type: ignore
                        [idx for idx, _ in idxs_by_dataset])
                    for (_, idx), item in zip(idxs_by_dataset, dataset_items):
                        items[idx] = item
            return items  # type: ignore
        else:
            return [self.__getitem__(idx) for idx in indices]

    def __len__(self):
        return self._len


__all__ = ["ExceptionHandlingMapDataset", "LinearMapSubset", "TransformedMapDataset",
           "ShuffledMapDataset", "UnionMapDataset", "DatasetToIterableDataset"]
